use std::{collections::HashMap, fmt::Display};

static INPUT: &str = include_str!("input");

fn main() {
    println!("part 1: {}", part1());
    println!("part 2: {}", part2());
}

struct Grid {
    data: Vec<Vec<char>>,
}

impl From<&str> for Grid {
    fn from(value: &str) -> Self {
        let data = value.lines().map(|line| line.chars().collect()).collect();
        Self { data }
    }
}

#[derive(Clone, Copy)]
struct Number {
    value: usize,
    location: Point,
    length: usize,
}

#[derive(Clone, Copy, Hash, PartialEq, Eq)]
struct Symbol {
    sym: char,
    location: Point,
}

#[derive(Clone, Copy, Hash, PartialEq, Eq)]
struct Point {
    x: usize,
    y: usize,
}

impl Display for Number {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{} at {},{}",
            self.value, self.location.x, self.location.y
        )
    }
}

impl Grid {
    fn get(&self, x: usize, y: usize) -> Option<char> {
        self.data.get(y).and_then(|row| row.get(x)).copied()
    }

    fn read_numbers(&self) -> Vec<Number> {
        self.data
            .iter()
            .enumerate()
            .flat_map(|(iy, line)| {
                let mut buf = String::new();
                let mut nums_on_line = Vec::new();
                let mut x = 0;
                for (ix, c) in line.iter().enumerate() {
                    if c.is_ascii_digit() {
                        if buf.is_empty() {
                            x = ix;
                        }
                        buf.push(*c);
                    } else if !buf.is_empty() {
                        let length = buf.len();
                        let value = buf.parse().unwrap();
                        buf = "".into();
                        nums_on_line.push(Number {
                            value,
                            length,
                            location: Point { x, y: iy },
                        });
                    }
                }
                if !buf.is_empty() {
                    let length = buf.len();
                    let value = buf.parse().unwrap();
                    nums_on_line.push(Number {
                        value,
                        length,
                        location: Point { x, y: iy },
                    });
                }
                nums_on_line
            })
            .collect()
    }

    fn get_adjacent_points(&self, num: &Number) -> Vec<Point> {
        let mut out = Vec::new();

        let x_min = num.location.x.saturating_sub(1);
        let x_max = num.location.x + num.length;
        let y_min = num.location.y.saturating_sub(1);
        let y_max = num.location.y + 1;

        for x in x_min..=x_max {
            for y in y_min..=y_max {
                out.push(Point { x, y })
            }
        }

        out
    }

    fn has_adjacent_symbol(&self, num: &Number) -> bool {
        self.get_adjacent_points(num)
            .into_iter()
            .filter_map(|p| self.get(p.x, p.y))
            .any(|c| c != '.' && !c.is_ascii_digit())
    }

    fn adjacent_symbols(&self, num: &Number) -> Vec<Symbol> {
        self.get_adjacent_points(num)
            .into_iter()
            .filter_map(|p| {
                self.get(p.x, p.y)
                    .filter(|c| *c != '.' && !c.is_ascii_digit())
                    .map(|c| Symbol {
                        sym: c,
                        location: p,
                    })
            })
            .collect()
    }
}

fn part1() -> usize {
    let grid = Grid::from(INPUT);
    grid.read_numbers()
        .into_iter()
        .filter(|num| grid.has_adjacent_symbol(num))
        .map(|num| num.value)
        .sum()
}

fn part2() -> usize {
    let grid = Grid::from(INPUT);

    let mut map: HashMap<Symbol, Vec<Number>> = HashMap::new();

    grid.read_numbers()
        .into_iter()
        .map(|num| (num, grid.adjacent_symbols(&num)))
        .filter(|(_, syms)| syms.iter().any(|sym| sym.sym == '*'))
        .map(|(num, syms)| {
            (
                syms.into_iter()
                    .filter(|s| s.sym == '*')
                    .collect::<Vec<_>>()[0],
                num,
            )
        })
        .for_each(|(sym, num)| {
            if let Some(v) = map.get_mut(&sym) {
                v.push(num);
            } else {
                map.insert(sym, vec![num]);
            }
        });

    map.into_values()
        .filter(|v| v.len() == 2)
        .map(|v| v.into_iter().map(|n| n.value).product::<usize>())
        .sum()
}
